/**
 * Copyright (c) 2008, Jose Enrique Bolanos, Jorge Villalobos
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *  * Neither the name of Jose Enrique Bolanos, Jorge Villalobos nor the names
 *    of its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 **/

Components.utils.import("resource://firefm/fmCommon.js");
Components.utils.import("resource://firefm/fmEntities.js");
Components.utils.import("resource://firefm/fmPlayerInitializer.js");
Components.utils.import("resource://firefm/fmPlayer.js");
Components.utils.import("resource://firefm/fmStation.js");
Components.utils.import("resource://firefm/fmNotifier.js");
Components.utils.import("resource://firefm/fmHistory.js");
Components.utils.import("resource://firefm/fmRemote.js");
Components.utils.import("resource://firefm/fmFeeds.js");
Components.utils.import("resource://gre/modules/Microformats.js");
Components.utils.import("resource://firefm/hAudio.js");

// How much of the selected text to display (start station context menu item)
const FIREFM_SELECTION_TEXT_MAX_LENGTH = 20;

/**
 * FireFM chrome namespace. We need a separate one because this one is defined
 * per window.
 */
if (typeof(FireFMChrome) == 'undefined') {
  var FireFMChrome = {};
};

/**
 * Browser overlay controller. This is the entry point for most operations that
 * happen in the toolbar.
 */
FireFMChrome.BrowserOverlay = {
  /* Logger for this object. */
  _logger : null,
  /* String bundle in the overlay. */
  _bundle : null,

  /**
   * Initializes the object.
   */
  init : function() {
    this._logger = Log4Moz.Service.getLogger("FireFMChrome.BrowserOverlay");
    this._logger.level = Log4Moz.Level["All"];
    this._logger.debug("init");

    // get the string bundle.
    this._bundle = document.getElementById("firefm-string-bundle");

    // Add listeners
    document.getElementById("contentAreaContextMenu").addEventListener(
      "popupshowing",FireFMChrome.BrowserOverlay.prepareContextMenu, false);

    FireFMChrome.UIState.init();
    FireFMChrome.Feeds.init();
    FireFM.startupDone = true;
  },

  /**
   * Unloads the object.
   */
  uninit : function() {
    this._logger.debug("uninit");

    // Remove listeners
    document.getElementById("contentAreaContextMenu").removeEventListener(
      "popupshowing",FireFMChrome.BrowserOverlay.prepareContextMenu);

    FireFMChrome.UIState.uninit();
  },

  /**
   * Prepares the Fire.fm menu items located in the content area context menu
   * before it is displayed.
   * @param aEvent The event object associated with this event.
   */
  prepareContextMenu : function(aEvent) {
    FireFMChrome.BrowserOverlay._logger.debug("prepareContextMenu");

    let miStartStation =
      document.getElementById("firefm-context-menu-start-station");
    let selection =
      document.commandDispatcher.focusedWindow.getSelection().toString();

    if (0 == selection.length) {
      miStartStation.hidden = true;

    } else {
      miStartStation.setAttribute("selection", selection);

      // Crop selection
      if (FIREFM_SELECTION_TEXT_MAX_LENGTH < selection.length) {
        selection =
          selection.substring(0, FIREFM_SELECTION_TEXT_MAX_LENGTH) + "...";
      }

      miStartStation.setAttribute("label",
        FireFMChrome.BrowserOverlay._bundle.getFormattedString(
          "firefm.context.play.label", [selection]));

      miStartStation.hidden = false;
    }
  },

  /**
   * Opens the Last.fm home page.
   * @param aEvent the event that triggered this action.
   */
  openHome : function(aEvent) {
    this._logger.debug("openHome");
    window.openUILink(FireFM.Remote.URL_HOME, aEvent);
  },

  /**
   * Opens the login or logout page for Last.fm, depending the user's current
   * state.
   * @param aEvent the event that triggered this action.
   */
  loginLogout : function(aEvent) {
    this._logger.debug("loginLogout");

    let url;

    if (aEvent.target.hasAttribute("loggedin")) {
      url = FireFM.Remote.URL_LOGOUT;
    } else {
      url = FireFM.Remote.URL_LOGIN;
    }

    window.openUILink(url, aEvent);
  },

  /**
   * Displays a dialog that asks the user to enter an artist's name to start a
   * station. The name is then verified against Last.fm.
   * @param aEvent the event that triggered this action.
   */
  startStation : function(aEvent) {
    this._logger.debug("startStation");

    let promptService =
      Cc["@mozilla.org/embedcomp/prompt-service;1"].
        getService(Ci.nsIPromptService);

    if (window.navigator.onLine &&
        FireFM.Player.status == FireFM.Player.STATUS_READY) {
      let artist = { value: "" };
      let userOK =
        promptService.prompt(
          window, this._bundle.getString("firefm.startAStation.label"),
          this._bundle.getString("firefm.typeArtist.label"), artist, null, {});

      if (userOK && (0 < artist.value.length)) {
        this.verifyArtist(artist.value);
      }
    } else {
      promptService.alert(
        window, this._bundle.getString("firefm.startAStation.label"),
        this._bundle.getString("firefm.offline.label"));
    }
  },

  /**
   * Verifies the artist's name against Last.FM, and loads the station if the
   * name is correct. Otherwise it opens a search page with the given search
   * string.
   * @param aArtist The name of the artist.
   */
  verifyArtist : function(aArtist) {
    this._logger.debug("verifyArtist");

    let that = this;

    FireFM.obsService.notifyObservers(
      null, FireFM.Station.TOPIC_STATION_SEARCHING, aArtist);
    FireFM.Remote.verifyArtist(
      aArtist, function(aResult) { that._verifyArtistLoad(aResult) });
  },

  /**
   * Load handler for the verify artist request.
   * @param aResult the resulting object from the request. See
   * FireFM.Remote.verifyArtist for more information.
   */
  _verifyArtistLoad : function(aResult) {
    this._logger.debug("_verifyArtistLoad");

    if (aResult.success) {
      FireFM.Station.setStation(aResult.result, FireFM.Station.TYPE_ARTIST);
      FireFM.Station.play();
      this._logger.debug("_verifyArtistLoad. Player loaded");
    } else {
      window.openUILinkIn(aResult.result, "tab");
      FireFM.obsService.notifyObservers(
        null, FireFM.Station.TOPIC_STATION_ERROR,
        FireFM.Station.ERROR_NOT_FOUND);
    }
  },

  /**
   * Starts a station with the recommendations from Last.fm for the logged in
   * user.
   * @param aEvent the event that triggered this action.
   */
  startRecommendations : function(aEvent) {
    this._logger.debug("startRecommendations");
    FireFM.Station.setStation(
      FireFM.Remote.userName, FireFM.Station.TYPE_RECOMMENDED);
    FireFM.Station.play();
  },

  /**
   * Starts the station for a friend or neighbor.
   * @param aEvent the event that triggered this action.
   */
  startUserStation : function(aEvent) {
    this._logger.debug("startUserStation");

    let userName = aEvent.originalTarget.getAttribute("label");

    FireFM.Station.setStation(userName, FireFM.Station.TYPE_USER);
    FireFM.Station.play();
  },

  /**
   * Starts the station for a top artist.
   * @param aEvent the event that triggered this action.
   */
  startArtistStation : function(aEvent) {
    this._logger.debug("startArtistStation");

    let artist = aEvent.originalTarget.getAttribute("label");

    this.verifyArtist(artist);
  },

  /**
   * Starts a recently played station.
   * @param aEvent the event that triggered this action.
   */
  startRecentStation : function(aEvent) {
    this._logger.debug("startRecentStation");

    let menuItem = aEvent.originalTarget;
    let stationId = menuItem.getAttribute("fmstationid");
    let stationType = parseInt(menuItem.getAttribute("fmstationtype"), 10);

    FireFM.Station.setStation(stationId, stationType);
    FireFM.Station.play();
  },

  /**
   * Plays or stops playback on a station.
   * @param aEvent the event that triggered this action.
   */
  playStop : function(aEvent) {
    this._logger.debug("playStop");

    if (aEvent.target.hasAttribute("playing")) {
      FireFM.Station.stop();
    } else {
      FireFM.Station.play();
    }
  },

  /**
   * Skips to the next track in the playlist.
   * @param aEvent the event that triggered this action.
   */
  skip : function(aEvent) {
    this._logger.debug("skip");
    FireFM.Station.skip();
  },

  /**
   * Sets the volume of the player.
   * @param aLevel The volume level to set.
   * @param aPersist True if the volume preference must be set. False otherwise.
   */
  setVolume : function(aLevel, aPersist) {
    // XXX: there is no logging here for performance reasons.
    try {
      FireFM.Player.setVolume(aLevel);
    } catch(e) {
      this._logger.error("setVolume. Error:\n" + e);
    }

    if (aPersist) {
      FireFMChrome.UIState.volumePref.value = aLevel;
    }
  },

  /**
   * Marks the currently played track as 'loved'.
   * @param aEvent the event that triggered this action.
   */
  loveTrack : function(aEvent) {
    this._logger.debug("loveTrack");
    // this is done to immediately disable the button and prevent multiple
    // submissions.
    FireFM.obsService.notifyObservers(
      null, FireFM.Remote.TOPIC_TRACK_LOVED, null);
    FireFM.Remote.loveTrack();
  },

  /**
   * Marks the currently played track as 'banned' and skips to the next one.
   * @param aEvent the event that triggered this action.
   */
  banTrack : function(aEvent) {
    this._logger.debug("banTrack");
    FireFM.Remote.banTrack();
    this.skip(aEvent);
  },

  /**
   * Fills the recent station menu popup with the most recently played stations.
   * @param aEvent the event that triggered this action.
   */
  fillRecentStations : function(aEvent) {
    this._logger.debug("fillRecentStations");

    let popup = aEvent.target;
    let recent = FireFM.History.stationHistory;
    let recentCount = recent.length;
    let menuItem;
    let station;

    // clear the popup.
    while (null != popup.firstChild) {
      popup.removeChild(popup.firstChild);
    }

    if (0 < recentCount) {
      for (let i = 0; i < recentCount; i++) {
        station = recent[i];
        menuItem = document.createElement("menuitem");
        menuItem.setAttribute("label",  station.title);
        menuItem.setAttribute("fmstationid",  station.id);
        menuItem.setAttribute("fmstationtype", station.type);

        if (FireFM.Station.TYPE_RECOMMENDED == station.type) {
          menuItem.setAttribute("observes", "firefm-logged-in-broadcaster");
        }

        popup.appendChild(menuItem);
      }
    } else {
      menuItem = document.createElement("menuitem");
      menuItem.setAttribute(
        "label",  this._bundle.getString("firefm.emptyMenu.label"));
      menuItem.setAttribute("disabled", true);
      popup.appendChild(menuItem);
    }
  },

  /**
   * Opens a page related to the currently playing track.
   * @param aEvent the event that triggered this action.
   */
  openTrackPage : function(aEvent) {
    this._logger.debug("openTrackPage");
    window.openUILink(aEvent.target.getAttribute("fmurl"), aEvent);
  },

  /**
   * Opens the preferences window.
   */
  openPreferences : function() {
    this._logger.debug("openPreferences");

    window.open(
      "chrome://firefm/content/fmPreferencesWindow.xul",
      "firefm-preferences-window",
      "chrome,centerscreen,titlebar,toolbar,resizable=no");
  },

  /**
   * Loads the hAudio microformats found in the current document into the given
   * menu popup.
   * @param aMenuPopup The menu popup to be loaded with the hAudio microformats.
   */
  loadAudioMicroformats : function(aMenuPopup) {
    this._logger.debug("loadAudioMicroformats");

    let audios = Microformats.get("hAudio", content.document);
    let item;

    while (aMenuPopup.firstChild) {
      aMenuPopup.removeChild(aMenuPopup.firstChild);
    }

    if (0 < audios.length) {
      for (var i = 0; i < audios.length; i++) {
        if (typeof(audios[i].title) != 'undefined' &&
            0 < audios[i].title.length) {
          item = document.createElement("menuitem");
          item.setAttribute("label", audios[i].title);
          item.setAttribute("oncommand",
            "FireFMChrome.BrowserOverlay.findAudioElement(event)");
          aMenuPopup.appendChild(item);
        }
      }
    }

    if (!aMenuPopup.hasChildNodes()) {
      item = document.createElement("menuitem");
      item.setAttribute(
        "label", this._bundle.getString("firefm.emptyMenu.label"));
      item.setAttribute("disabled", true);
      aMenuPopup.appendChild(item);
    }
  },

  /**
   * Called on oncommand events of hAudio microformat menu items. Finds the page
   * element associated with the menu item's microformat and scrolls it into
   * view.
   * @param aEvent The event that triggered this action.
   */
  findAudioElement : function(aEvent) {
    this._logger.debug("findAudioElement");

    var title = aEvent.originalTarget.getAttribute("label");
    var elements = content.document.getElementsByClassName("title");
    var element;
    var decoration;

    for (var i = 0; i < elements.length; i++) {
      element = elements[i];

      if (title == element.textContent) {
        element.scrollIntoView();

        decoration = element.style.textDecoration;
        element.style.textDecoration = "blink";

        window.setTimeout(function() {
          element.style.textDecoration = decoration;
        }, 2500);

        break;
      }
    }
  }
};

window.addEventListener(
  "load", function() { FireFMChrome.BrowserOverlay.init(); }, false);
window.addEventListener(
  "unload", function() { FireFMChrome.BrowserOverlay.uninit(); }, false);
