/**
 * Copyright (c) 2008, Jose Enrique Bolanos, Jorge Villalobos
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *  * Neither the name of Jose Enrique Bolanos, Jorge Villalobos nor the names
 *    of its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 **/

/**
 * This object manages the state of all buttons and other elements that are
 * visible to the user. We have complex state transitions, such as station
 * loading, actual playback, login/logout, and online/offline.
 */
FireFMChrome.UIState = {
  /* Observer topics this object listens to. */
  _TOPICS :
    [ FireFM.Station.TOPIC_STATION_SET,
      FireFM.Station.TOPIC_STATION_SEARCHING,
      FireFM.Station.TOPIC_STATION_LOADING,
      FireFM.Station.TOPIC_STATION_OPENING,
      FireFM.Station.TOPIC_STATION_STOPPING,
      FireFM.Station.TOPIC_STATION_ERROR,
      FireFM.Player.TOPIC_TRACK_LOADED,
      FireFM.Player.TOPIC_PROGRESS_CHANGED,
      FireFM.Playlist.TOPIC_SKIPS_CHANGED,
      FireFM.Remote.TOPIC_USER_AUTHENTICATION,
      FireFM.Remote.TOPIC_TRACK_LOVED ],

  /* Timeout before running autoplay. */
  _AUTOPLAY_TIMEOUT : 100, // 100 miliseconds.
  /* Timeout before showing the player error UI. */
  _PLAYER_ERROR_TIMEOUT : 2 * 1000, // 2 seconds.

  /* Logger for this object. */
  _logger : null,
  /* String bundle in the overlay. */
  _bundle : null,
  /* Online/offline application state. */
  _onLine : false,
  /* Indicates if the player is ready. */
  _isPlayerReady : false,
  /* Indicates if the station is playing or not. */
  _stationPlaying : false,
  /* The currently logged in user. null if logged out. */
  _currentUser : null,
  /* Indicates whether the current track has been loved or not. */
  _loved : false,
  /* Volume preference. */
  _volumePref : null,
  /* Reference to the volume change listener. */
  _volumeListener : null,
  /* The number of skips left for the user. */
  _skipsLeft : 100,
  /* The Now Playing broadcaster, which is frequently used. */
  _nowPlaying : null,
  /* The time left menu item, which is frequently used. */
  _timeLeftMenu : null,

  /**
   * Initializes the object.
   */
  init : function() {
    let that = this;
    let playerStatus;

    this._logger = Log4Moz.Service.getLogger("FireFMChrome.UIState");
    this._logger.level = Log4Moz.Level["All"];
    this._logger.debug("init");

    // get the string bundle.
    this._bundle = document.getElementById("firefm-string-bundle");

    // get the Now Playing broadcaster and time left menu item.
    this._nowPlaying =
      document.getElementById("firefm-now-playing-broadcaster");
    this._timeLeftMenu =
      document.getElementById("firefm-menu-playing-time-left");

    // set all preference objects and listeners.
    this._volumePref =
      Application.prefs.get(FireFM.PREF_BRANCH + "volumeLevel");
    this._volumeListener =
      function(aEvent) { that._setVolume(that._volumePref.value); };

    // add observers.
    for (let i = 0; i < this._TOPICS.length; i++) {
      FireFM.obsService.addObserver(this, this._TOPICS[i], false);
    }

    this._volumePref.events.addListener("change", this._volumeListener);

    // gather state information.
    playerStatus = FireFM.Player.status;
    this._onLine = window.navigator.onLine;
    this._isPlayerReady = (FireFM.Player.STATUS_READY == playerStatus);
    this._stationPlaying = FireFM.Player.isPlaying;
    this._currentUser = FireFM.Remote.userName;
    this._setVolume(this._volumePref.value);

    // this only needs to be run for the first window being opened.
    if (!FireFM.startupDone) {
      // show an error message in case the player didn't load appropriately.
      if (!this._isPlayerReady) {
        window.setTimeout(
          function() { that._handlePlayerError(playerStatus); },
          this._PLAYER_ERROR_TIMEOUT);
      }
      // this needs to be run before autoplay so that stations that depend on
      // the logged in state work.
      FireFM.Remote.checkLoggedInState();

      if (0 < FireFM.History.stationHistory.length) {
        let lastStation = FireFM.History.stationHistory[0];

        try {
          FireFM.Station.setStation(lastStation.id, lastStation.type);
          window.setTimeout(
            function() { that._setPlayerReady(); }, this._AUTOPLAY_TIMEOUT);
        } catch (e) {
          this._logger.error(
            "init. Error setting station from preference:\n" + e);
        }
      }
    }

    this._setPreferencesCommand();
    // finally, refresh the state of the whole UI.
    this.refreshState();

    // set the now playing info for new windows.
    if (FireFM.Player.isPlaying) {
      this._setNowPlayingInfo(FireFM.Playlist.currentTrack);
    }
  },

  /**
   * Unloads the object.
   */
  uninit : function() {
    this._logger.debug("uninit");
    // remove observers.
    for (let i = 0; i < this._TOPICS.length; i++) {
      FireFM.obsService.removeObserver(this, this._TOPICS[i]);
    }

    this._volumePref.events.removeListener("change", this._volumeListener);
  },

  /**
   * Sets the player for autoplay if it is ready and the preferences are set.
   */
  _setPlayerReady : function() {
    this._logger.trace("_setPlayerReady");

    if (this._isPlayerReady) {
      let autoplay = Application.prefs.get(FireFM.PREF_BRANCH + "autoplay");

      try {
        // set the volume.
        FireFMChrome.BrowserOverlay.setVolume(this._volumePref.value, false)

        // autoplay if set by the user.
        if (autoplay && autoplay.value) {
          FireFM.Station.play();
        }
      } catch (e) {
        this._logger.error("_setPlayerReady. Error setting up player:\n" + e);
      }
    }
  },

  /**
   * Sets the label and accesskey for the preferences command. This needs
   * special attention because the label is platform-specific.
   */
  _setPreferencesCommand : function() {
    this._logger.trace("_setPreferencesCommand");

    let prefCmd = document.getElementById("firefm-preferences-cmd");
    let os = FireFM.getOperatingSystem();

    if ((FireFM.OS_WINDOWS == os) || (FireFM.OS_WINDOWS_VISTA == os)) {
      prefCmd.setAttribute(
        "label", this._bundle.getString("firefm.options.label"));
      prefCmd.setAttribute(
        "accesskey", this._bundle.getString("firefm.options.accesskey"));
    } else {
      prefCmd.setAttribute(
        "label", this._bundle.getString("firefm.optionsUnix.label"));
      prefCmd.setAttribute(
        "accesskey", this._bundle.getString("firefm.optionsUnix.accesskey"));
    }
  },

  /**
   * Refreshes the state of the UI components.
   */
  refreshState : function() {
    this._logger.debug("refreshState");
    this._refreshLoginBroadcaster();
    this._refreshLoginButton();
    this._refreshStationBroadcaster();
    this._refreshPlayButton();
    this._refreshSkipButton();
    this._refreshVolumeBroadcaster();
    this._refreshLoveButton();
    this._refreshBanButton();
    this._refreshRecentStationMenu();
    this._refreshNowPlayingMenu();
  },

  /**
   * Refreshes the state of the login broadcaster.
   */
  _refreshLoginBroadcaster : function() {
    this._logger.trace("_refreshLoginBroadcaster");

    let loginBroadcaster =
      document.getElementById("firefm-logged-in-broadcaster");

    if (this._onLine && this._isPlayerReady && (null != this._currentUser)) {
      loginBroadcaster.removeAttribute("disabled");
    } else {
      loginBroadcaster.setAttribute("disabled", true);
    }
  },

  /**
   * Refreshes the state of the login / logout button.
   */
  _refreshLoginButton : function() {
    this._logger.trace("_refreshLoginButton");

    let loginLogoutCmd = document.getElementById("firefm-login-logout-cmd");

    loginLogoutCmd.removeAttribute("disabled");

    if (this._onLine) {
      if (null != this._currentUser) {
        let logoutStr =
          this._bundle.getFormattedString(
            "firefm.logout.label", [ this._currentUser ]);

        loginLogoutCmd.setAttribute("loggedin", true);
        loginLogoutCmd.setAttribute("label", logoutStr);
        loginLogoutCmd.setAttribute(
          "accesskey", this._bundle.getString("firefm.logout.accesskey"));
        loginLogoutCmd.tooltipText = logoutStr;
      } else {
        let loginStr = this._bundle.getString("firefm.login.label");

        loginLogoutCmd.removeAttribute("loggedin");
        loginLogoutCmd.setAttribute("label", loginStr);
        loginLogoutCmd.setAttribute(
          "accesskey", this._bundle.getString("firefm.login.accesskey"));
        loginLogoutCmd.tooltipText = loginStr;
      }
    } else {
      loginLogoutCmd.removeAttribute("loggedin");
      loginLogoutCmd.setAttribute(
        "label", this._bundle.getString("firefm.login.label"));
      loginLogoutCmd.setAttribute(
        "accesskey", this._bundle.getString("firefm.login.accesskey"));
      loginLogoutCmd.tooltipText =
        this._bundle.getString(
          (!this._onLine ? "firefm.offline.label" :
           "firefm.playerNotLoaded.label"));
      loginLogoutCmd.setAttribute("disabled", true);
    }
  },

  /**
   * Refreshes the state of the station broadcaster.
   */
  _refreshStationBroadcaster : function() {
    this._logger.trace("_refreshStationBroadcaster");

    let stationBroadcaster =
      document.getElementById("firefm-station-menu-broadcaster");

    if (this._isPlayerReady) {
      stationBroadcaster.removeAttribute("disabled");
    } else {
      stationBroadcaster.setAttribute("disabled", true);
    }
  },

  /**
   * Refreshes the state of the play / stop button.
   */
  _refreshPlayButton : function() {
    this._logger.trace("_refreshPlayButton");

    let playStopCmd = document.getElementById("firefm-play-stop-cmd");

    playStopCmd.removeAttribute("disabled");

    if (this._onLine && this._isPlayerReady) {
      let station = FireFM.Station.station;

      if (this._stationPlaying) {
        playStopCmd.setAttribute("playing", true);
        playStopCmd.setAttribute(
          "label", this._bundle.getString("firefm.stop.label"));
        playStopCmd.setAttribute(
          "accesskey", this._bundle.getString("firefm.stop.accesskey"));
        playStopCmd.tooltipText =
          this._bundle.getFormattedString(
            "firefm.stop.tooltip", [ station.title ]);
      } else {
        playStopCmd.removeAttribute("playing");
        playStopCmd.setAttribute(
          "label", this._bundle.getString("firefm.play.label"));
        playStopCmd.setAttribute(
          "accesskey", this._bundle.getString("firefm.play.accesskey"));

        if (null != station) {
          playStopCmd.tooltipText =
            this._bundle.getFormattedString(
              "firefm.play.tooltip", [ station.title ]);
        } else {
          playStopCmd.tooltipText =
            this._bundle.getString("firefm.playDefault.tooltip");
          playStopCmd.setAttribute("disabled", true);
        }
      }
    } else {
      playStopCmd.removeAttribute("playing");
      playStopCmd.setAttribute(
        "label", this._bundle.getString("firefm.play.label"));
      playStopCmd.setAttribute(
        "accesskey", this._bundle.getString("firefm.play.accesskey"));
      playStopCmd.tooltipText =
        this._bundle.getString(
          (!this._onLine ? "firefm.offline.label" :
           "firefm.playerNotLoaded.label"));
      playStopCmd.setAttribute("disabled", true);
    }
  },

  /**
   * Refreshes the state of the skip button.
   */
  _refreshSkipButton : function() {
    this._logger.trace("_refreshSkipButton");

    let skipCmd = document.getElementById("firefm-skip-cmd");

    if (this._onLine && this._isPlayerReady) {
      if (this._stationPlaying) {
        if (0 < this._skipsLeft) {
          skipCmd.removeAttribute("disabled");
          skipCmd.tooltipText = this._bundle.getString("firefm.skip.tooltip");
        } else {
          skipCmd.setAttribute("disabled", true);
          skipCmd.tooltipText =
            this._bundle.getString("firefm.skipNoLeft.tooltip");
        }
      } else {
        skipCmd.setAttribute("disabled", true);
        skipCmd.tooltipText = this._bundle.getString("firefm.skip.tooltip");
      }
    } else {
      skipCmd.setAttribute("disabled", true);
      skipCmd.tooltipText = this._bundle.getString("firefm.skip.tooltip");
    }
  },

  /**
   * Refreshes the state of the volume broadcaster.
   */
  _refreshVolumeBroadcaster : function() {
    this._logger.trace("_refreshVolumeBroadcaster");

    let volumeBroadcaster =
      document.getElementById("firefm-volume-broadcaster");

    if (this._isPlayerReady) {
      volumeBroadcaster.removeAttribute("disabled");
    } else {
      volumeBroadcaster.setAttribute("disabled", true);
    }
  },

  /**
   * Refreshes the state of the love track button.
   */
  _refreshLoveButton : function() {
    this._logger.trace("_refreshLoveButton");
    // oh yeah, the love command :P
    let loveCmd = document.getElementById("firefm-love-cmd");

    loveCmd.setAttribute("disabled", true);

    if (this._onLine && this._isPlayerReady && (null != this._currentUser)) {
      if (this._stationPlaying) {
        if (!this._loved) {
          if (!this._lovedError) {
            loveCmd.tooltipText = this._bundle.getString("firefm.love.tooltip");
          } else {
            loveCmd.tooltipText =
              this._bundle.getString("firefm.lovedError.tooltip");
          }

          loveCmd.removeAttribute("disabled");
        } else {
          loveCmd.tooltipText =
            this._bundle.getString("firefm.alreadyLoved.tooltip");
        }
      } else {
        loveCmd.tooltipText = this._bundle.getString("firefm.love.tooltip");
      }
    } else {
      loveCmd.tooltipText = this._bundle.getString("firefm.needLogin.label");
    }
  },

  /**
   * Refreshes the state of the ban track button.
   */
  _refreshBanButton : function() {
    this._logger.trace("_refreshBanButton");

    let banCmd = document.getElementById("firefm-ban-cmd");

    banCmd.setAttribute("disabled", true);

    if (this._onLine && this._isPlayerReady && (null != this._currentUser)) {
      banCmd.tooltipText = this._bundle.getString("firefm.ban.tooltip");

      if (this._stationPlaying) {
        banCmd.removeAttribute("disabled");
      }
    } else {
      banCmd.tooltipText = this._bundle.getString("firefm.needLogin.label");
    }
  },

  /**
   * Refreshes the state of the recent stations menu.
   */
  _refreshRecentStationMenu : function() {
    this._logger.trace("_refreshRecentStationMenu");

    let recentCmd = document.getElementById("firefm-recent-station-cmd");

    if (this._onLine && this._isPlayerReady) {
      recentCmd.removeAttribute("disabled");
    } else {
      recentCmd.setAttribute("disabled", true);
    }
  },

  /**
   * Refreshes the state of the Now Playing menu.
   */
  _refreshNowPlayingMenu : function() {
    this._logger.trace("_refreshNowPlayingMenu");

    let nowPlaying = document.getElementById("firefm-menu-now-playing");

    if (null != nowPlaying) {
      if (this._onLine && this._isPlayerReady && this._stationPlaying) {
        nowPlaying.removeAttribute("disabled");
      } else {
        nowPlaying.setAttribute("disabled", true);
      }
    }
  },

  /**
   * Sets the Now Playing information with the given track.
   * @param aTrack the track to use to set the information.
   */
  _setNowPlayingInfo : function(aTrack) {
    this._logger.trace("_setNowPlayingInfo");

    let infoMenuItem = document.getElementById("firefm-menu-playing-info");
    let trackMenuItem = document.getElementById("firefm-menu-playing-track");
    let artistMenuItem = document.getElementById("firefm-menu-playing-artist");
    let albumMenuItem = document.getElementById("firefm-menu-playing-album");
    let imagePath;

    if (aTrack.imagePath.length > 0) {
      imagePath =  aTrack.imagePath;
    } else {
      imagePath = "chrome://firefm/skin/logo100.png";
    }

    this._nowPlaying.setAttribute("tracktitle", aTrack.title);
    this._nowPlaying.setAttribute("trackimage", imagePath);
    this._nowPlaying.setAttribute("trackurl", aTrack.trackURL);
    this._nowPlaying.setAttribute("artist", aTrack.artist);
    this._nowPlaying.setAttribute("artisturl", aTrack.artistURL);
    this._nowPlaying.setAttribute("album", aTrack.albumTitle);
    this._nowPlaying.setAttribute("albumurl", aTrack.albumURL);
    this._nowPlaying.removeAttribute("message");
    this._nowPlaying.removeAttribute("progress");

    if (null != infoMenuItem) {
      let trackString = aTrack.title + " - " + aTrack.artist;

      infoMenuItem.setAttribute("label", trackString);
      infoMenuItem.setAttribute("image", imagePath);
      infoMenuItem.setAttribute("tooltiptext", trackString);
    }

    if (null != trackMenuItem) {
      if (null != aTrack.trackURL) {
        trackMenuItem.setAttribute("fmurl", aTrack.trackURL);
        trackMenuItem.removeAttribute("disabled");
      } else {
        trackMenuItem.setAttribute("disabled", true);
      }
    }

    if (null != artistMenuItem) {
      if (null != aTrack.artistURL) {
        artistMenuItem.setAttribute("fmurl", aTrack.artistURL);
        artistMenuItem.removeAttribute("disabled");
      } else {
        artistMenuItem.setAttribute("disabled", true);
      }
    }

    if (null != albumMenuItem) {
      if (null != aTrack.albumURL) {
        albumMenuItem.setAttribute("fmurl", aTrack.albumURL);
        albumMenuItem.removeAttribute("disabled");
      } else {
        albumMenuItem.setAttribute("disabled", true);
      }
    }
  },

  /**
   * Clearss the Now Playing information.
   */
  _clearNowPlayingInfo : function() {
    this._logger.trace("_clearNowPlayingInfo");
    this._nowPlaying.removeAttribute("trackimage");
    this._nowPlaying.removeAttribute("tracktitle");
    this._nowPlaying.removeAttribute("trackurl");
    this._nowPlaying.removeAttribute("artist");
    this._nowPlaying.removeAttribute("artisturl");
    this._nowPlaying.removeAttribute("album");
    this._nowPlaying.removeAttribute("albumurl");
  },

  /**
   * Sets the current online state.
   * @param aIsOnline whether the application is online or not. true if the
   * application is online, false otherwise.
   */
  _setOnlineState : function(aIsOnline) {
    this._logger.trace("_setOnlineState");
    this._onLine = aIsOnline;

    if (!aIsOnline && this._stationPlaying) {
      FireFM.Station.stop();
    }

    this.refreshState();
  },

  /**
   * Gets the volume Preference object.
   * @return the volume Preference object.
   */
  get volumePref() {
    // XXX: there is no logging here for performance reasons.
    return this._volumePref;
  },

  /**
   * Sets the volume in the volume scale.
   * @param aLevel the volume level in a scale from 0 to 100.
   */
  _setVolume : function(aLevel) {
    // XXX: there is no logging here for performance reasons.
    let volumeBroadcaster =
      document.getElementById("firefm-volume-broadcaster");
    let levelAttr = "high";

    if (aLevel == 1) {
      levelAttr = "zero";
    } else if (aLevel <= 50) {
      levelAttr = "low";
    }

    volumeBroadcaster.setAttribute("value", aLevel);
    volumeBroadcaster.setAttribute("volume", levelAttr);
  },

  /**
   * Sets the given message in the Now Playing observer.
   * @param aMessageKey the message key that holds the message. If null, the
   * message is cleared.
   * @param aParameter (optional) the parameter added to the message.
   */
  _setNowPlayingMessage : function(aMessageKey, aParameter) {
    this._logger.trace("_setNowPlayingMessage");

    if (null != aMessageKey) {
      if (aParameter) {
        this._nowPlaying.setAttribute(
          "message",
          this._bundle.getFormattedString(aMessageKey, [ aParameter ]));
      } else {
        this._nowPlaying.setAttribute(
          "message", this._bundle.getString(aMessageKey));
      }
    } else {
      this._nowPlaying.removeAttribute("message");
    }
  },

  /**
   * Displays a message or window to the user informing about the Player error
   * that occurred.
   * @param aPlayerStatus the player error status.
   */
  _handlePlayerError : function(aPlayerStatus) {
    this._logger.error("_handlePlayerError. Status: " + aPlayerStatus);

    this._setNowPlayingMessage("firefm.playerNotLoaded.label");

    if (FireFM.Player.STATUS_PLUGIN_MISSING == aPlayerStatus) {
      // show a notification in the browser area, telling the user that the
      // Flash plugin is not present or not up to date.
      let nb = gBrowser.getNotificationBox();
      let installButton = new Object();

      installButton.label =
        this._bundle.getString("firefm.pluginNotification.installPlugin.label");
      installButton.accessKey =
        this._bundle.getString(
          "firefm.pluginNotification.installPlugin.accesskey");
      installButton.popup = null;
      installButton.callback =
        function() {
          openUILinkIn("http://www.adobe.com/go/getflashplayer", "tab"); };

      nb.appendNotification(
        this._bundle.getString("firefm.pluginNotification.label"),
        "firefm-plugin-notification", "chrome://firefm/skin/logo32.png",
        nb.PRIORITY_WARNING_HIGH, [ installButton ] );

    } else {
      // open an error reporting window that lets the user send us the error
      // that caused the player not to load.

      window.open(
        "chrome://firefm/content/fmErrorDialog.xul",
        "firefm-error-dialog",
       "chrome,dialog,centerscreen").focus();
    }
  },

  /**
   * Observes topic notifications.
   * @param aSubject The object that experienced the change.
   * @param aTopic The topic being observed.
   * @param aData The data relating to the change.
   */
  observe : function(aSubject, aTopic, aData) {
    // XXX: there is no logging here for performance purposes.
    switch (aTopic) {
      case FireFM.Station.TOPIC_STATION_SEARCHING:
        this._setNowPlayingMessage(
          "firefm.searching.label", FireFM.decodeFMString(aData));
        break;
      case FireFM.Station.TOPIC_STATION_SET:
        this.refreshState();
        break;
      case FireFM.Station.TOPIC_STATION_LOADING:
        this.refreshState();
        this._setNowPlayingMessage(
          "firefm.loading.label",
          decodeURIComponent(aSubject.wrappedJSObject.id));
        break;
      case FireFM.Station.TOPIC_STATION_OPENING:
        this._stationPlaying = true;
        this.refreshState();
        this._setNowPlayingMessage(
          "firefm.opening.label",
          decodeURIComponent(aSubject.wrappedJSObject.title));
        break;
      case FireFM.Station.TOPIC_STATION_STOPPING:
        this._stationPlaying = false;
        this.refreshState();
        this._clearNowPlayingInfo();
        this._setNowPlayingMessage(
          "firefm.stopped.label",
          decodeURIComponent(aSubject.wrappedJSObject.title));
        break;
      case FireFM.Station.TOPIC_STATION_ERROR:
        this._setNowPlayingMessage("firefm.error." + aData + ".label");

        if (this._stationPlaying) {
          this._stationPlaying = false;
          this.refreshState();

          if (FireFM.Station.ERROR_NOT_FOUND == parseInt(aData)) {
            let that = this;
            // clear the message, so that the current station can continue
            // showing its information.
            window.setTimeout(
              function() { that._setNowPlayingMessage(null); }, 5000);
          }
        }

        break;
      case FireFM.Player.TOPIC_TRACK_LOADED:
        this._loved = false;
        this._lovedError = false;
        this.refreshState();
        this._setNowPlayingInfo(aSubject.wrappedJSObject);
        break;
      case FireFM.Player.TOPIC_PROGRESS_CHANGED:
        let remaining = FireFM.Player.remainingTime;

        this._nowPlaying.setAttribute("progress", aData);
        this._nowPlaying.setAttribute("remainingtime", "-" + remaining);
        this._nowPlaying.setAttribute("isbuffering", FireFM.Player.isBuffering);

        if (null != this._timeLeftMenu) {
          this._timeLeftMenu.setAttribute(
            "label",
            this._bundle.getFormattedString(
              "firefm.timeLeft.label", [ remaining ]));
        }
        break;
      case FireFM.Playlist.TOPIC_SKIPS_CHANGED:
        this._skipsLeft = parseInt(aData, 10);
        this._refreshSkipButton();
        break;
      case FireFM.Remote.TOPIC_USER_AUTHENTICATION:
        if (null != aData) {
          this._logger.debug("observe. Logged in as: " + aData);
          this._currentUser = aData;
          this.refreshState();
        } else {
          this._logger.debug("observe. Logged out.");
          this._currentUser = null;
          this.refreshState();
        }

        break;
      case FireFM.Remote.TOPIC_TRACK_LOVED:
        if (null == aData) {
          this._loved = true;
          this.refreshState();
        } else if ("false" == aData) {
          this._lovedError = true;
          this._loved = false;
          this.refreshState();
        }

        break;
    }
  }
};

window.addEventListener(
  "online", function() { FireFMChrome.UIState._setOnlineState(true); }, false);
window.addEventListener(
  "offline", function() { FireFMChrome.UIState._setOnlineState(false); },
  false);
