/**
 * Copyright (c) 2008, Jose Enrique Bolanos, Jorge Villalobos
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *  * Neither the name of Jose Enrique Bolanos, Jorge Villalobos nor the names
 *    of its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 **/

var EXPORTED_SYMBOLS = [ "FireFM.Initializer" ];

const Cc = Components.classes;
const Ci = Components.interfaces;
const Ce = Components.Exception;

Components.utils.import("resource://firefm/fmCommon.js");
Components.utils.import("resource://firefm/fmPlayer.js");

/**
 * FireFM Player Initializer. Performs all the flash player intialization tasks.
 */
if (typeof(FireFM.PlayerInitializer) == 'undefined') {
FireFM.PlayerInitializer = {

  /* Logger for this object */
  _logger : null,
  /* Holds the exception that may have been thrown during initialization */
  _initializationException : null,

  /**
   * Initializes the object.
   */
  init : function() {
    this._logger = Log4Moz.Service.getLogger("FireFM.PlayerInitializer");
    this._logger.level = Log4Moz.Level["All"];
    this._logger.debug("init");

    this._initializeFlashPlayer();
  },

  /**
   * Obtains the exception that may have been thrown during the flash player
   * initialization. Null if no exception was thrown.
   */
  get initializationException() {
    this._logger.debug("initializationException[get]");
    return this._initializationException;
  },

  /**
   * Initializes the flash object needed to play mp3 files. Steps:
   * 1. The flash plugin is detected.
   * 2. The flash player URL is calculated.
   * 3. The flash player permission file is installed.
   * 4. The flash player object is injected in the hidden window.
   * After the initialization tasks are complete, FireFM.Player is initialized
   * with the results.
   */
  _initializeFlashPlayer : function() {
    this._logger.trace("_initializeFlashPlayer");

    let flashURL = null;
    let flashObject = null;
    let status = null;

    try {
      // test error line. Comment out on releases!
      //null.hello;

      if (!this._detectFlashPlugin()) {

        status = FireFM.Player.STATUS_PLUGIN_MISSING;
        this._logger.warn("Flash plugin missing");

      } else {
        flashURL = this._getFlashURL();
        this._installPermissionFile(flashURL);

        flashObject = this._injectFlash(flashURL);
        status = FireFM.Player.STATUS_READY;

        this._logger.info("Flash player initialized successfully");
      }
    } catch (e) {
      status = FireFM.Player.STATUS_LOAD_ERROR;
      this._initializationException = e;

      this._logger.fatal("Flash player initialization failed:\n" + e);
    }

    FireFM.Player.initializePlayer(flashObject, status);
  },

  /**
   * Detects if the flash player plugin is installed, enabled, and of the
   * required version.
   * @return True of the plugin meets the requirements, false otherwise.
   */
  _detectFlashPlugin : function() {
    this._logger.trace("_detectFlashPlugin");

    // Regular expression to obtain the flash plugin version from its
    // description.
    const REGEX_FLASH_VERSION = /([0-9]+)[^0-9]/gi;
    // Minimum flash version supported
    const MIN_FLASH_VERSION = 8;

    let isValidPlugin = false;

    let plugin =
      Cc["@mozilla.org/appshell/appShellService;1"].
        getService(Ci.nsIAppShellService).hiddenDOMWindow.
          navigator.mimeTypes["application/x-shockwave-flash"];

    if (plugin && plugin.enabledPlugin && plugin.enabledPlugin.description) {

      let match = REGEX_FLASH_VERSION.exec(plugin.enabledPlugin.description);
      if (match && match.length > 0) {
        isValidPlugin = (MIN_FLASH_VERSION <= parseInt(match[1]));
      }
    }

    return isValidPlugin;
  },

  /**
   * Obtains the URL of the flash player, located in the extension's default
   * folder.
   * @return The URL of the flash player.
   */
  _getFlashURL : function() {
    this._logger.trace("_getFlashURL");

    // UUID of the Fire.fm extension
    const EXTENSION_UUID = "{6F0976E6-26F3-4AFE-BBEC-9E99E27E4DF3}";
    // Flash file name
    const FLASH_FILE_NAME = "firefm.swf";

    let extensionManager =
      Cc["@mozilla.org/extensions/manager;1"].
        getService(Ci.nsIExtensionManager);

    let flashFile =
      extensionManager.
        getInstallLocation(EXTENSION_UUID).getItemLocation(EXTENSION_UUID);
    flashFile.append("defaults");
    flashFile.append(FLASH_FILE_NAME);

    let flashURL =
      Cc["@mozilla.org/network/protocol;1?name=file"].
        getService(Ci.nsIFileProtocolHandler).getURLSpecFromFile(flashFile);

    return flashURL;
  },

  /**
   * Installs the permission file required for the embedded flash object to
   * connect to the Internet. If the permission file and entry already exist
   * then the file is not altered.
   * @param aFlashURL The URL of the flash player, used in the permission file.
   */
  _installPermissionFile : function(aFlashURL) {
    this._logger.trace("_installPermissionFile");

    let permissionFile = this._getPermissionFile();
    let data =
      "resource://gre/res/hiddenWindow.html\n" +
      "chrome://browser/content/hiddenWindow.xul\n" +
      aFlashURL;

    let foStream =
      Cc["@mozilla.org/network/file-output-stream;1"].
        createInstance(Ci.nsIFileOutputStream);

    // write, create, truncate
    foStream.init(permissionFile, 0x02 | 0x08 | 0x20, 0666, 0);
    foStream.write(data, data.length);
    foStream.close();

    this._logger.debug("Flash permission file written");
  },

  /**
   * Obtains a reference to the flash permission file.
   * @return The file permission file reference.
   */
  _getPermissionFile : function() {
    this._logger.trace("_getPermissionFile");

    // Name of the permission file
    const PERMISSION_FILE_NAME = "firefm.cfg";
    // Strings passed to the directory service to obtain the application data
    // folder on each OS
    const LOCATIONS = [
      // Windows
      "AppData",
      // Mac
      "UsrPrfs",
      // Unix
      "Home"
    ];
    // Directory path to the permission file for each OS
    const DIR_PATH = [
      // Windows
      ["Macromedia", "Flash Player", "#Security", "FlashPlayerTrust"],
      // Mac
      ["Macromedia", "Flash Player", "#Security", "FlashPlayerTrust"],
      // Unix
      [".macromedia", "Flash_Player", "#Security", "FlashPlayerTrust"]
    ];

    let osIndex = 0;
    switch (FireFM.getOperatingSystem()) {
      case FireFM.OS_MAC:   osIndex = 1; break;
      case FireFM.OS_LINUX: osIndex = 2; break;
    }

    let file =
      Cc["@mozilla.org/file/directory_service;1"].
        getService(Ci.nsIProperties).
          get(LOCATIONS[osIndex], Ci.nsIFile);

    // Move towards target path
    for (let i = 0; i < DIR_PATH[osIndex].length; i++) {
      file.append(DIR_PATH[osIndex][i]);

      if (!file.exists() || !file.isDirectory()) {
        file.create(Ci.nsIFile.DIRECTORY_TYPE, 0777);
      }
    }
    file.append(PERMISSION_FILE_NAME);

    return file;
  },

  /**
   * Injects the embedded flash player object needed to play mp3 files in the
   * hidden window.
   * @param aFlashURL The URL from where to load the flash player.
   */
  _injectFlash : function(aFlashURL) {
    this._logger.trace("_injectFlash");

    let doc =
      Cc["@mozilla.org/appshell/appShellService;1"].
        getService(Ci.nsIAppShellService).hiddenDOMWindow.document;

    let playerObj =
      doc.createElementNS("http://www.w3.org/1999/xhtml", "embed");

    playerObj.setAttribute("src", aFlashURL);
    playerObj.setAttribute("type", "application/x-shockwave-flash");
    playerObj.setAttribute("id", "firefm-player");
    playerObj.setAttribute("name", "firefm-player");
    playerObj.setAttribute("FlashVars", "id='firefm-player'");
    playerObj.setAttribute("allowScriptAccess", "always");
    playerObj.setAttribute("quality", "high");

    doc.documentElement.appendChild(playerObj);

    return playerObj;
  }
};}

/**
 * FireFM.PlayerInitializer constructor.
 */
(function() {
  this.init();
}).apply(FireFM.PlayerInitializer);
